/* 12-18 October 2020, by R.Harmsen

   Demonstration of all the parameters of a Soundblaster/
   Adlib/Yamaha OPL chip. Pressing keys on the keyboard
   changes parameters, and immediately sounds the result.

   Target platform: MSDos / DosBox

   15-17 October 2020: Add chord mode
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include "fmchip.h"
#include "relspeed.h"
#include "showintv.h"

static struct OPL_ValuesStruct
{
   int FM_mode;
   int Feedback;
   double freq;
   int multiplier_carrier;
   int multiplier_modulator;
   int level_carrier;
   int level_modulator;
   int KSL_carrier;
   int KSL_modulator;
   int WS_carrier;
   int WS_modulator;

   int KSR_carrier;
   int sustain_carrier;
   int attack_rate_carrier;
   int decay_rate_carrier;
   int sustain_lvl_carrier;
   int release_rate_carrier;

   int KSR_modulator;
   int sustain_modulator;
   int attack_rate_modulator;
   int decay_rate_modulator;
   int sustain_lvl_modulator;
   int release_rate_modulator;

   int tremolo_carrier;   /* 0 = OFF, 1 = weak, 2 = strong */
   int tremolo_modulator; /* 0 = OFF, 1 = ON */
   int vibrato_carrier;   /* 0 = OFF, 1 = weak, 2 = strong */
   int vibrato_modulator; /* 0 = OFF, 1 = ON */

} OPL_values = {1, 0, 440, 1, 1, 0, 32, 1, 1, 0, 0,
   1, 1, 9 /* attack */, 1 /* decay */, 15 /* sustain */, 1  /* release */,
   1, 1, 9 /* attack */, 1 /* decay */, 15 /* sustain */, 1  /* release */,
   0, 0, 0, 0};

static chord_mode;
#define notes_in_chord 5
/* Frequency factors are relative to the unison. */
static struct chord_factor_struct
{
   double freq_ratio;
   short int is_on;
   struct interv_struct *match;
} chord_factors[notes_in_chord] =
/* Default chord is dominant ninth, but seventh and ninth are off */
{
   {     1, 1, NULL /* unison  */ },
   { 6/5.0, 1, NULL /* third   */ },
   { 3/2.0, 1, NULL /* fifth   */ },
   { 9/5.0, 0, NULL /* seventh */ },
   { 9/4.0, 0, NULL /* ninth   */ },
};

static void get_values (unsigned scan);
static void set_values (void);
static void Snap2OctaveDivision (int equalparts);
static void display (void);

static void adjust_chord  (int scan);
static void play_chord    (void);
static void display_chord (void);

static void notes_off (void);

static unsigned scankeyb (void);

/**************************************************************************

 *************************************************************************/
int main (void)
{
   unsigned scan;

   init_fmchip();
   set_nts(0);
   set_values();
   note_on(1);
   display();

   /* Loop ends on escape key */
   while ((scan = scankeyb()) != 0x01)
   {
      if (scan == 0)
         continue;

      #ifdef DEBUG
      fprintf(stdout, "Scancode = %2x\n", scan);
      #endif

      if (scan == 0x3c)/* Function key F2 */
      {
         chord_mode = !chord_mode;
         if (chord_mode)
            set_values();
      }

      if (scan == 0x4f) /* End key */
         notes_off();
      else if (scan == 0x39) /* space bar */
         chord_mode ? play_chord() : note_on(1);
      else
      {
         if (chord_mode)
         {
            adjust_chord(scan);
            play_chord();
            display_chord();
         }
         else
         {
            get_values(scan);
            set_values();
            note_on(1);
            display();
         }
      }
   }

   return 0;
}

void get_values (unsigned scan)
{
   int up = -1; /* -1 is down, 1 is up */

   switch (scan)
   {
      case 0x02: /* digit 1 */
         /* Toggle synthesizer mode */
         OPL_values.FM_mode = !OPL_values.FM_mode;
         break;
      case 0x10: /* q */
         if (++(OPL_values.Feedback) > 7)
            OPL_values.Feedback = 0;
         break;

      case 0x1e: /* a */
         if (++(OPL_values.WS_carrier) > 7)
            OPL_values.WS_carrier = 0;
         break;
      case 0x2c: /* z */
         if (++(OPL_values.WS_modulator) > 7)
            OPL_values.WS_modulator = 0;
         break;

      case 0x03: /* 2 */
         up = 1;
      case 0x11: /* w */
         OPL_values.multiplier_carrier += up;
         if (OPL_values.multiplier_carrier > 15)
            OPL_values.multiplier_carrier = 0;
         else if (OPL_values.multiplier_carrier < 0)
            OPL_values.multiplier_carrier = 15;
         break;
      case 0x1f: /* z */
         up = 1;
      case 0x2d: /* x */
         OPL_values.multiplier_modulator += up;
         if (OPL_values.multiplier_modulator > 15)
            OPL_values.multiplier_modulator = 0;
         else if (OPL_values.multiplier_modulator < 0)
            OPL_values.multiplier_modulator = 15;
         break;

      case 0x12: /* e */
         up = 1;
      case 0x04: /* 3 */
         OPL_values.level_carrier += up;
         if (OPL_values.level_carrier > 63)
            OPL_values.level_carrier = 0;
         else if (OPL_values.level_carrier < 0)
         {
            /* Increase KSL, then rollover */
            if (++OPL_values.KSL_carrier > 3)
               OPL_values.KSL_carrier = 0;
            OPL_values.level_carrier = 63;
         }
         break;
      case 0x2e: /* c */
         up = 1;
      case 0x20: /* d */
         OPL_values.level_modulator += up;
         if (OPL_values.level_modulator > 63)
            OPL_values.level_modulator = 0;
         else if (OPL_values.level_modulator < 0)
         {
            /* Increase KSL, then rollover */
            if (++OPL_values.KSL_modulator > 3)
               OPL_values.KSL_modulator = 0;
            OPL_values.level_modulator = 63;
         }
         break;

      /* Frequency down/up in 100 cent steps (1/12 of an octave)*/
      case 0x05: /* 4 */
         up = 1;
      case 0x13: /* r */
         OPL_values.freq *= pow(2, up * 1/12.0);
         OPL_values.freq = value2freq(freq2value(OPL_values.freq));
         break;

      /* Frequency down/up in 22.6 cent steps (1/53 of an octave) */
      case 0x21: /* f */
         up = 1;
      case 0x2f: /* v */
         OPL_values.freq *= pow(2, up * 1/53.0);
         /* Recalculate to OPL chip resulution (10 bits) */
         OPL_values.freq = value2freq(freq2value(OPL_values.freq));
         break;

      /* Snap frequency to nearest multiple of 100 cents */
      case 0x06: /* 5 */
            Snap2OctaveDivision(12);
         break;
      /* Snap frequency to nearest multiple of 50 cents */
      case 0x14: /* t */
            Snap2OctaveDivision(24);
         break;

      /* Frequency down/up by minimal Soundblaster/Adlib/OPL-chip stepsize,
         in 10-bit resolution */
      case 0x22: /* g */
         up = 1;
      case 0x30: /* b */
         {
            int value = freq2value(OPL_values.freq) & 0x3ff;

            OPL_values.freq *= (value + up) / (double)value;
            /* Recalculate to OPL chip resulution (10 bits) */
            OPL_values.freq = value2freq(freq2value(OPL_values.freq));
         }
         break;

      /* Envelope carrier */
      case 0x07: /* digit 6 */
         up = 1;
      case 0x15: /* y */
         OPL_values.attack_rate_carrier += up;
         if (OPL_values.attack_rate_carrier > 15)
            OPL_values.attack_rate_carrier  = 0;
         else if (OPL_values.attack_rate_carrier  < 0)
            OPL_values.attack_rate_carrier  = 15;
         break;

      case 0x08: /* digit 7 */
         up = 1;
      case 0x16: /* u */
         OPL_values.decay_rate_carrier += up;
         if (OPL_values.decay_rate_carrier > 15)
            OPL_values.decay_rate_carrier  = 0;
         else if (OPL_values.decay_rate_carrier  < 0)
            OPL_values.decay_rate_carrier  = 15;
         break;

      case 0x09: /* digit 8 */
         up = 1;
      case 0x17: /* i */
         OPL_values.sustain_lvl_carrier += up;
         if (OPL_values.sustain_lvl_carrier > 15)
            OPL_values.sustain_lvl_carrier  = 0;
         else if (OPL_values.sustain_lvl_carrier  < 0)
            OPL_values.sustain_lvl_carrier  = 15;
         break;

      case 0x0a: /* digit 9 */
         up = 1;
      case 0x18: /* o */
         OPL_values.release_rate_carrier += up;
         if (OPL_values.release_rate_carrier > 15)
            OPL_values.release_rate_carrier  = 0;
         else if (OPL_values.release_rate_carrier  < 0)
            OPL_values.release_rate_carrier  = 15;
         break;

      /* Envelope modulator */
      case 0x23: /* h */
         up = 1;
      case 0x31: /* n */
         OPL_values.attack_rate_modulator += up;
         if (OPL_values.attack_rate_modulator > 15)
            OPL_values.attack_rate_modulator  = 0;
         else if (OPL_values.attack_rate_modulator  < 0)
            OPL_values.attack_rate_modulator  = 15;
         break;

      case 0x24: /* j */
         up = 1;
      case 0x32: /* m */
         OPL_values.decay_rate_modulator += up;
         if (OPL_values.decay_rate_modulator > 15)
            OPL_values.decay_rate_modulator  = 0;
         else if (OPL_values.decay_rate_modulator  < 0)
            OPL_values.decay_rate_modulator  = 15;
         break;

      case 0x25: /* k */
         up = 1;
      case 0x33: /* , */
         OPL_values.sustain_lvl_modulator += up;
         if (OPL_values.sustain_lvl_modulator > 15)
            OPL_values.sustain_lvl_modulator  = 0;
         else if (OPL_values.sustain_lvl_modulator  < 0)
            OPL_values.sustain_lvl_modulator  = 15;
         break;

      case 0x26: /* l */
         up = 1;
      case 0x34: /* . */
         OPL_values.release_rate_modulator += up;
         if (OPL_values.release_rate_modulator > 15)
            OPL_values.release_rate_modulator  = 0;
         else if (OPL_values.release_rate_modulator  < 0)
            OPL_values.release_rate_modulator  = 15;
         break;

      case 0x0b: /* 0 */
         OPL_values.KSR_carrier = !OPL_values.KSR_carrier;
         break;
      case 0x19: /* p */
         OPL_values.sustain_carrier = !OPL_values.sustain_carrier;
         break;
      case 0x27: /* ; */
         OPL_values.KSR_modulator = !OPL_values.KSR_modulator;
         break;
      case 0x35: /* / */
         OPL_values.sustain_modulator = !OPL_values.sustain_modulator;
         break;

      case 0x0c: /* - */
         if (++(OPL_values.tremolo_carrier) > 2)
            OPL_values.tremolo_carrier = 0;
         break;
      case 0x1a: /* [ */
         if (++(OPL_values.tremolo_modulator) > 2)
            OPL_values.tremolo_modulator = 0;
         break;
      case 0x0d: /* + */
         if (++(OPL_values.vibrato_carrier) > 2)
            OPL_values.vibrato_carrier = 0;
         break;
      case 0x1b: /* ] */
         if (++(OPL_values.vibrato_modulator) > 2)
            OPL_values.vibrato_modulator = 0;
         break;
   }
}

void Snap2OctaveDivision (int equalparts)
{
   double steps =
      fmod(log(OPL_values.freq / (440/32.0)) / log(2) * equalparts,
           equalparts);
   double intsteps;

   steps = modf(steps + 0.5, &intsteps);
   OPL_values.freq /= pow(2, (steps - 0.5) / equalparts);
}

void set_values (void)
{
   int i;

   for (i = 0; i < (chord_mode ? notes_in_chord : 1); i++)
   {
      set_synth_mode (i + 1, OPL_values.FM_mode ? FREQMODU_SYNTH : ADDITIVE_SYNTH);
      set_fm_feedback(i + 1, OPL_values.Feedback);
      set_ws         (i + 1, CARRIER  , OPL_values.WS_carrier);
      set_ws         (i + 1, MODULATOR, OPL_values.WS_modulator);
      set_multiplier (i + 1, CARRIER  , OPL_values.multiplier_carrier);
      set_multiplier (i + 1, MODULATOR, OPL_values.multiplier_modulator);
      set_level      (i + 1, CARRIER  , OPL_values.level_carrier);
      set_level      (i + 1, MODULATOR, OPL_values.level_modulator);
      set_ksl        (i + 1, CARRIER  , OPL_values.KSL_carrier);
      set_ksl        (i + 1, MODULATOR, OPL_values.KSL_modulator);

      set_frequency_f(i + 1, OPL_values.freq);

      set_ksr(1,CARRIER, OPL_values.KSR_carrier);
      set_envelope(i + 1, CARRIER, OPL_values.sustain_carrier,
         OPL_values.attack_rate_carrier,
         OPL_values.decay_rate_carrier,
         OPL_values.sustain_lvl_carrier,
         OPL_values.release_rate_carrier);

      set_ksr(i + 1, MODULATOR, OPL_values.KSR_modulator);
      set_envelope(i + 1, MODULATOR, OPL_values.sustain_modulator,
         OPL_values.attack_rate_modulator,
         OPL_values.decay_rate_modulator,
         OPL_values.sustain_lvl_modulator,
         OPL_values.release_rate_modulator);

      set_am_vib(i + 1, CARRIER,
         OPL_values.tremolo_carrier  , OPL_values.vibrato_carrier);
      set_am_vib(i + 1, MODULATOR,
         OPL_values.tremolo_modulator, OPL_values.vibrato_modulator);
      /* There is only a single value setting for all settings */
      seti_am_vib (OPL_values.tremolo_carrier == 2 || OPL_values.tremolo_modulator == 2,
                   OPL_values.vibrato_carrier == 2 || OPL_values.vibrato_modulator == 2);
   }
}

void display (void)
{
   fprintf(stdout, "escape: stop and quit, space: note(s) on, end: note(s) off, F2: chord mode\n");
   fprintf(stdout, "1\tToggle additive/freq.modul.\t%s\n",
      OPL_values.FM_mode ? "Frequency modulation" : "Additive synthesis");
   fprintf(stdout, "q\tFreq.modulation feedback   \t%2d\n", OPL_values.Feedback);

   fprintf(stdout, "a\tCarrier   waveform sel (WS)\t%2d\n", OPL_values.WS_carrier);
   fprintf(stdout, "z\tModulator waveform sel (WS)\t%2d\n", OPL_values.WS_modulator);

   fprintf(stdout, "w/2\tCarrier multiplier       \t%2d\n", OPL_values.multiplier_carrier);
   fprintf(stdout, "x/s\tModulator multiplier     \t%2d\n", OPL_values.multiplier_modulator);

   fprintf(stdout, "e/3\tCarrier level            \t%2d\n", OPL_values.level_carrier);
   fprintf(stdout, "c/d\tModulator level          \t%2d\n", OPL_values.level_modulator);
   fprintf(stdout, "0>63\tCarrier KSL             \t%2d\n", OPL_values.KSL_carrier);
   fprintf(stdout, "0>63\tModulator KSL           \t%2d\n", OPL_values.KSL_modulator);

   fprintf(stdout, "r/4\tFrequency (1/12 octave)\n");
   fprintf(stdout, "v/f\tFrequency (1/53 octave) \t%8.3f Hz %4d %8.3f cent\n",
      OPL_values.freq, freq2value(OPL_values.freq) & 0x3ff,
      fmod(log(OPL_values.freq / (440/32.0)) / log(2) * 1200, 1200));
   fprintf(stdout, "5/t\tSnap to nearest 100/50 cents\n");
   fprintf(stdout, "b/g\tFrequency (1 OPL chip step; 10-bits resolution)\n");

   fprintf(stdout, "0p6789/yuio\tCarrier:\nKSR %s sustain %s attack %2d decay %2d sust.lvl. %2d release %2d\n",
      OPL_values.KSR_carrier ? "y" : "n",
      OPL_values.sustain_carrier ? "y" : "n",
      OPL_values.attack_rate_carrier,
      OPL_values.decay_rate_carrier,
      OPL_values.sustain_lvl_carrier,
      OPL_values.release_rate_carrier);

   fprintf(stdout, ";/hjkl/nm,.\tModulator:\nKSR %s sustain %s attack %2d, decay %2d, sust.lvl. %2d, release %2d\n",
      OPL_values.KSR_modulator ? "y" : "n",
      OPL_values.sustain_modulator ? "y" : "n",
      OPL_values.attack_rate_modulator,
      OPL_values.decay_rate_modulator,
      OPL_values.sustain_lvl_modulator,
      OPL_values.release_rate_modulator);

   fprintf(stdout, "-+\tCarrier   tremolo %d vibrato\t%2d\n",
      OPL_values.tremolo_carrier  , OPL_values.vibrato_carrier);
   fprintf(stdout, "[]\tModulator tremolo %d vibrato\t%2d\n",
      OPL_values.tremolo_modulator, OPL_values.vibrato_modulator);
}

static void adjust_chord (int scan)
{
   int note_in_chord;
   int down = 1; /* 1 = up, -1 = down */

   switch (scan)
   {
      /* keys 3, 4, 5, 6, 7 */
      case 0x04: case 0x05: case 0x06: case 0x07: case 0x08:
         note_in_chord = 4;
         break;

      /* keys e, r, t, y, u */
      case 0x12: case 0x13: case 0x14: case 0x15: case 0x16:
         note_in_chord = 3;
         break;

      /* keys d, f, g, h, j */
      case 0x20: case 0x21: case 0x22: case 0x23: case 0x24:
         note_in_chord = 2;
         break;

      /* keys c, v, b, n, m */
      case 0x2e: case 0x2f: case 0x30: case 0x31: case 0x32:
         note_in_chord = 1;
         break;

      /* key x (= toggle on/off, of unison */
      case 0x2d:
         note_in_chord = 0;
         break;
   }

   switch (scan)
   {
      /* Enable/disable notes of the chord */
      /* keys 3, e, d, c, x */
      case 0x04: case 0x12: case 0x20: case 0x2e: case 0x2d:
         chord_factors[note_in_chord].is_on =
        !chord_factors[note_in_chord].is_on;
         break;

      /* Adjust frequency in steps equal to SB/OPL resolution (10 bits) */
      /* keys 4, r, f, v */
      case 0x05: case 0x13: case 0x21: case 0x2f:
         down = -1;
      /* keys 5, t, g, b */
      case 0x06: case 0x14: case 0x22: case 0x30:
         {
            int value = freq2value(
               OPL_values.freq * chord_factors[note_in_chord].freq_ratio)
               & 0x3ff;

            chord_factors[note_in_chord].freq_ratio *=
               (value + down) / (double)value;
         }
         break;

      /* Adjust frequency in steps equal to 1/53 of an octave */
      /* keys 6, y, h, n */
      case 0x07: case 0x15: case 0x23: case 0x31:
         down = -1;
      /* keys 7, u, j, m */
      case 0x08: case 0x16: case 0x24: case 0x32:
      {
         double value, freq;

         chord_factors[note_in_chord].freq_ratio *= pow(2, down * 1/53.0);

         freq = OPL_values.freq * chord_factors[note_in_chord].freq_ratio;
         /* Snap to what a frequency that can really be produced
            by the Yamaha OPL chip. */
         freq = value2freq(freq2value(freq));
         /* Adjust the ratio accordingly. */
         chord_factors[note_in_chord].freq_ratio = freq / OPL_values.freq;
         break;
      }
   }

}

static void display_chord (void)
{
   int i;
   double freq, cents;
   struct interv_struct *prv = NULL, *nxt = NULL;

   fprintf(stdout, "\n\n\n");

   for (i = notes_in_chord - 1; i >= 0; i--)
   {
      if (chord_factors[i].is_on)
      {
         freq = OPL_values.freq * chord_factors[i].freq_ratio;
         cents = log(chord_factors[i].freq_ratio) / log(2) * 1200;
         chord_factors[i].match =
            showintv(cents, chord_factors[i].match, &prv, &nxt);
         fprintf(stdout, "%*s %8.3f %s\n", 50, "", nxt->cents, nxt->text);
      }
      else
         fprintf(stdout, "\n");

      switch (i)
      {
         case 4: /* ninth */
            fprintf(stdout, "3/45/67 ");
            break;
         case 3: /* seventh */
            fprintf(stdout, "e/rt/yu ");
            break;
         case 2: /* fifth */
            fprintf(stdout, "d/fg/hj ");
            break;
         case 1: /* third */
            fprintf(stdout, "c/vb/nm ");
            break;
         case 0: /* unison; not adjustable */
            fprintf(stdout, "x/fixed ");
            break;
      }
      if (chord_factors[i].is_on)
      {
         fprintf(stdout, "%8.3f Hz %4d %8.3f cent (%6.3f) => %8.3f %s\n",
            freq, freq2value(freq) & 0x3ff, cents,
            fmod(cents, 1200) - chord_factors[i].match->cents,
            chord_factors[i].match->cents,
            chord_factors[i].match->text);
         fprintf(stdout, "%*s %8.3f %s\n\n", 50, "", prv->cents, prv->text);
      }
      else
      {
         fprintf(stdout, " OFF\n\n\n");
      }
   }
}

static void play_chord (void)
{
   int i;

   for (i = 0; i < notes_in_chord; i++)
   {
      if (chord_factors[i].is_on)
      {
         set_frequency_f(i + 1, OPL_values.freq * chord_factors[i].freq_ratio);
         note_on(i + 1);
      }
   }
}

static void notes_off (void)
{
   int i;

   for (i = 0; i < notes_in_chord; i++)
      note_off(i + 1);
}

/**************************************************************************
   Look if a key has been pressed, and if so, return its scancode. If
   not, it returns 0.
 *************************************************************************/
#include <dos.h>
static unsigned scankeyb (void)
{
   union REGS r;

   r.h.ah = 0x01;
   int86(0x16, &r, &r);

   if (!(r.x.flags & 0x40))
   {
      /* Key pressed, return it */
      r.h.ah = 0x00;
      int86(0x16, &r, &r);
      return r.h.ah;
   }
   return 0;
}
